import itertools
import torch
import torch.nn as nn
import torch.nn.functional as F


class StackedDilatedConv(nn.Module):
    def __init__(
        self, in_channels, out_channels, kernel_size, dilation_rates, groups, dim=2
    ):
        """
        A unified class for Stacked Dilated Convolutions supporting both 2D and 3D.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            kernel_size (int or tuple): Size of the convolutional kernel.
            dilation_rates (list): List of dilation rates for the stacked convolutions.
            groups (int): Number of groups for grouped convolution.
            dim (int): Dimensionality of the convolution (2 for 2D, 3 for 3D).
        """
        super(StackedDilatedConv, self).__init__()
        assert in_channels == out_channels, "in_channels must be equal to out_channels"
        assert dim in [2, 3], "dim must be 2 (2D) or 3 (3D)"
        self.dim = dim
        self.dilation_rates = dilation_rates

        # Choose the appropriate convolutional layer based on the dimension
        Conv = nn.Conv2d if dim == 2 else nn.Conv3d
        self.base_conv = Conv(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            groups=groups,
        )
        self.reduction_conv = Conv(
            in_channels=len(dilation_rates) * out_channels,
            out_channels=out_channels,
            kernel_size=1,
            groups=groups,
        )

    def forward(self, x):
        outputs = []
        # Choose the appropriate functional convolution based on the dimension
        conv_func = F.conv2d if self.dim == 2 else F.conv3d

        for dilation_rate in self.dilation_rates:
            outputs.append(
                conv_func(
                    x,
                    self.base_conv.weight,
                    self.base_conv.bias,
                    padding=dilation_rate,
                    dilation=dilation_rate,
                    groups=self.base_conv.groups,
                )
            )

        outputs = [
            # Split the output into groups
            torch.split(out, out.shape[1] // self.base_conv.groups, dim=1)
            for out in outputs
        ]

        outputs = list(itertools.chain(*zip(*outputs)))
        outputs = torch.concat(outputs, dim=1)
        outputs = F.leaky_relu(outputs)

        return self.reduction_conv(outputs)



import torch
import torch.nn as nn
import torch.nn.functional as F
from ignite.utils import to_onehot
from pytorch_rdc_net.lovasz_losses import lovasz_hinge, lovasz_hinge3d


class AbstractInstanceEmbeddingLoss(nn.Module):
    def __init__(self, dim):
        super(AbstractInstanceEmbeddingLoss, self).__init__()
        self.dim = dim
        if self.dim == 2:
            self.lovasz_loss = lovasz_hinge
        elif self.dim == 3:
            self.lovasz_loss = lovasz_hinge3d
        else:
            raise ValueError("dim must be 2 or 3")

    def get_instance_size(self, gt_one_hot):
        return torch.sum(gt_one_hot, dim=tuple(range(1, self.dim + 1)), keepdim=True)

    def get_lower_triangular_values(self, y_lt, gt_one_hot, instance_size):
        instance_lower_triangular = gt_one_hot.unsqueeze(0) * y_lt.unsqueeze(1)
        lower_triangular_instance_mean = (
            torch.sum(
                instance_lower_triangular,
                dim=tuple(range(2, self.dim + 2)),
                keepdim=True,
            )
            / instance_size
        )

        lower_triangular_instance_variance = torch.sum(
            torch.pow(
                (instance_lower_triangular - lower_triangular_instance_mean.detach())
                * gt_one_hot,
                2,
            ),
            dim=tuple(range(2, self.dim + 2)),
            keepdim=True,
        ) / instance_size.unsqueeze(0)
        return lower_triangular_instance_mean, lower_triangular_instance_variance

    def compute_probabilities(self, y_emb, grid, gt_one_hot, instance_size, cov):
        centers_mean = (
            torch.sum(
                (gt_one_hot.unsqueeze(0) * grid.unsqueeze(1)),
                dim=tuple(range(2, self.dim + 2)),
                keepdim=True,
            )
            / instance_size
        )
        similar_embs = torch.square(y_emb.unsqueeze(1) - centers_mean)
        probs = []
        for i in range(similar_embs.shape[1]):
            s_e = similar_embs[:, i].unsqueeze(1)
            s_e = torch.moveaxis(s_e.view(self.dim, 1, -1), -1, 0)
            probs.append(
                torch.exp(
                    -1
                    * (torch.transpose(s_e, 1, 2) @ cov[i] @ s_e).view(
                        1, *similar_embs.shape[-self.dim :]
                    )
                )
            )
        return torch.cat(probs, dim=0)

    def forward(self, grid, y_embeddings, y_lower_triangular, y_true):
        losses = []
        lt_vars = []
        for y_emb, y_lt, gt_patch in zip(y_embeddings, y_lower_triangular, y_true):
            if torch.any(gt_patch > 0):
                gt_one_hot = to_onehot(
                    gt_patch, num_classes=int(torch.max(gt_patch).item() + 1)
                )[0, 1:]
                instance_size = self.get_instance_size(gt_one_hot=gt_one_hot)
                lt_mean, lt_var = self.get_lower_triangular_values(
                    y_lt=y_lt, gt_one_hot=gt_one_hot, instance_size=instance_size
                )
                cov = self.get_covariance_matrix(lt_mean)

                probabilities = self.compute_probabilities(
                    y_emb=y_emb,
                    grid=grid,
                    gt_one_hot=gt_one_hot,
                    instance_size=instance_size,
                    cov=cov,
                )

                losses.append(
                    self.lovasz_loss(probabilities * 2 - 1, gt_one_hot, per_image=False)
                )
                lt_vars.append(torch.mean(lt_var))

        if len(losses) > 0:
            return torch.mean(torch.stack(losses)), torch.mean(torch.stack(lt_vars))
        else:
            return torch.tensor(0.0), torch.tensor(0.0)

    def get_covariance_matrix(self, lt_mean):
        raise NotImplementedError("This method must be implemented in subclasses.")


class InstanceEmbeddingLoss2d(AbstractInstanceEmbeddingLoss):
    def __init__(self):
        super(InstanceEmbeddingLoss2d, self).__init__(dim=2)

    def get_covariance_matrix(self, lt_mean):
        lt_matrix = torch.zeros(
            (lt_mean.shape[1], 2, 2), device=lt_mean.device, dtype=lt_mean.dtype
        )
        lt_matrix[:, 0, 0] = F.softplus(lt_mean[0, :, 0, 0])
        lt_matrix[:, 1, 1] = F.softplus(lt_mean[1, :, 0, 0])
        lt_matrix[:, 1, 0] = lt_mean[2, :, 0, 0]
        return torch.bmm(lt_matrix, lt_matrix.transpose(1, 2))


class InstanceEmbeddingLoss3d(AbstractInstanceEmbeddingLoss):
    def __init__(self):
        super(InstanceEmbeddingLoss3d, self).__init__(dim=3)

    def get_covariance_matrix(self, lt_mean):
        lt_matrix = torch.zeros(
            (lt_mean.shape[1], 3, 3), device=lt_mean.device, dtype=lt_mean.dtype
        )
        lt_matrix[:, 0, 0] = F.softplus(lt_mean[0, :, 0, 0, 0])
        lt_matrix[:, 1, 1] = F.softplus(lt_mean[1, :, 0, 0, 0])
        lt_matrix[:, 2, 2] = F.softplus(lt_mean[2, :, 0, 0, 0])
        lt_matrix[:, 1, 0] = lt_mean[3, :, 0, 0, 0]
        lt_matrix[:, 2, 0] = lt_mean[4, :, 0, 0, 0]
        lt_matrix[:, 2, 1] = lt_mean[5, :, 0, 0, 0]
        return torch.bmm(lt_matrix, lt_matrix.transpose(1, 2))



"""
From [StarDist](https://github.com/stardist/stardist).

BSD 3-Clause License

Copyright (c) 2018-2024, Uwe Schmidt, Martin Weigert
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np
from tqdm import tqdm
from numba import jit
from scipy.optimize import linear_sum_assignment
from skimage.measure import regionprops
from collections import namedtuple

matching_criteria = dict()


def label_are_sequential(y):
    """returns true if y has only sequential labels from 1..."""
    labels = np.unique(y)
    return (set(labels) - {0}) == set(range(1, 1 + labels.max()))


def is_array_of_integers(y):
    return isinstance(y, np.ndarray) and np.issubdtype(y.dtype, np.integer)


def _check_label_array(y, name=None, check_sequential=False):
    err = ValueError(
        "{label} must be an array of {integers}.".format(
            label="labels" if name is None else name,
            integers=("sequential " if check_sequential else "")
            + "non-negative integers",
        )
    )
    is_array_of_integers(y) or err
    if len(y) == 0:
        return True
    if check_sequential:
        label_are_sequential(y) or err
    else:
        y.min() >= 0 or err
    return True


def label_overlap(x, y, check=True):
    if check:
        _check_label_array(x, "x", True)
        _check_label_array(y, "y", True)
        x.shape == y.shape or ValueError("x and y must have the same shape")
    return _label_overlap(x, y)


@jit(nopython=True)
def _label_overlap(x, y):
    x = x.ravel()
    y = y.ravel()
    overlap = np.zeros((1 + x.max(), 1 + y.max()), dtype=np.uint)
    for i in range(len(x)):
        overlap[x[i], y[i]] += 1
    return overlap


def _safe_divide(x, y, eps=1e-10):
    """computes a safe divide which returns 0 if y is zero"""
    if np.isscalar(x) and np.isscalar(y):
        return x / y if np.abs(y) > eps else 0.0
    else:
        out = np.zeros(np.broadcast(x, y).shape, np.float32)
        np.divide(x, y, out=out, where=np.abs(y) > eps)
        return out


def intersection_over_union(overlap):
    _check_label_array(overlap, "overlap")
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_pred = np.sum(overlap, axis=0, keepdims=True)
    n_pixels_true = np.sum(overlap, axis=1, keepdims=True)
    return _safe_divide(overlap, (n_pixels_pred + n_pixels_true - overlap))


matching_criteria["iou"] = intersection_over_union


def intersection_over_true(overlap):
    _check_label_array(overlap, "overlap")
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_true = np.sum(overlap, axis=1, keepdims=True)
    return _safe_divide(overlap, n_pixels_true)


matching_criteria["iot"] = intersection_over_true


def intersection_over_pred(overlap):
    _check_label_array(overlap, "overlap")
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_pred = np.sum(overlap, axis=0, keepdims=True)
    return _safe_divide(overlap, n_pixels_pred)


matching_criteria["iop"] = intersection_over_pred


def precision(tp, fp, fn):
    return tp / (tp + fp) if tp > 0 else 0


def recall(tp, fp, fn):
    return tp / (tp + fn) if tp > 0 else 0


def accuracy(tp, fp, fn):
    # also known as "average precision" (?)
    # -> https://www.kaggle.com/c/data-science-bowl-2018#evaluation
    return tp / (tp + fp + fn) if tp > 0 else 0


def f1(tp, fp, fn):
    # also known as "dice coefficient"
    return (2 * tp) / (2 * tp + fp + fn) if tp > 0 else 0


def matching(y_true, y_pred, thresh=0.5, criterion="iou", report_matches=False):
    """Calculate detection/instance segmentation metrics between ground truth and predicted label images.

    Currently, the following metrics are implemented:

    'fp', 'tp', 'fn', 'precision', 'recall', 'accuracy', 'f1', 'criterion', 'thresh', 'n_true', 'n_pred', 'mean_true_score', 'mean_matched_score', 'panoptic_quality'

    Corresponding objects of y_true and y_pred are counted as true positives (tp), false positives (fp), and false negatives (fn)
    whether their intersection over union (IoU) >= thresh (for criterion='iou', which can be changed)

    * mean_matched_score is the mean IoUs of matched true positives

    * mean_true_score is the mean IoUs of matched true positives but normalized by the total number of GT objects

    * panoptic_quality defined as in Eq. 1 of Kirillov et al. "Panoptic Segmentation", CVPR 2019

    Parameters
    ----------
    y_true: ndarray
        ground truth label image (integer valued)
    y_pred: ndarray
        predicted label image (integer valued)
    thresh: float
        threshold for matching criterion (default 0.5)
    criterion: string
        matching criterion (default IoU)
    report_matches: bool
        if True, additionally calculate matched_pairs and matched_scores (note, that this returns even gt-pred pairs whose scores are below  'thresh')

    Returns
    -------
    Matching object with different metrics as attributes

    Examples
    --------
    >>> y_true = np.zeros((100,100), np.uint16)
    >>> y_true[10:20,10:20] = 1
    >>> y_pred = np.roll(y_true,5,axis = 0)

    >>> stats = matching(y_true, y_pred)
    >>> print(stats)
    Matching(criterion='iou', thresh=0.5, fp=1, tp=0, fn=1, precision=0, recall=0, accuracy=0, f1=0, n_true=1, n_pred=1, mean_true_score=0.0, mean_matched_score=0.0, panoptic_quality=0.0)

    """
    _check_label_array(y_true, "y_true")
    _check_label_array(y_pred, "y_pred")
    y_true.shape == y_pred.shape or ValueError(
        "y_true ({y_true.shape}) and y_pred ({y_pred.shape}) have different shapes".format(
            y_true=y_true, y_pred=y_pred
        )
    )
    criterion in matching_criteria or ValueError(
        "Matching criterion '%s' not supported." % criterion
    )
    if thresh is None:
        thresh = 0
    thresh = float(thresh) if np.isscalar(thresh) else map(float, thresh)

    y_true, _, map_rev_true = relabel_sequential(y_true)
    y_pred, _, map_rev_pred = relabel_sequential(y_pred)

    overlap = label_overlap(y_true, y_pred, check=False)
    scores = matching_criteria[criterion](overlap)
    assert 0 <= np.min(scores) <= np.max(scores) <= 1

    # ignoring background
    scores = scores[1:, 1:]
    n_true, n_pred = scores.shape
    n_matched = min(n_true, n_pred)

    def _single(thr):
        # not_trivial = n_matched > 0 and np.any(scores >= thr)
        not_trivial = n_matched > 0
        if not_trivial:
            # compute optimal matching with scores as tie-breaker
            costs = -(scores >= thr).astype(float) - scores / (2 * n_matched)
            true_ind, pred_ind = linear_sum_assignment(costs)
            assert n_matched == len(true_ind) == len(pred_ind)
            match_ok = scores[true_ind, pred_ind] >= thr
            tp = np.count_nonzero(match_ok)
        else:
            tp = 0
        fp = n_pred - tp
        fn = n_true - tp
        # assert tp+fp == n_pred
        # assert tp+fn == n_true

        # the score sum over all matched objects (tp)
        sum_matched_score = (
            np.sum(scores[true_ind, pred_ind][match_ok]) if not_trivial else 0.0
        )

        # the score average over all matched objects (tp)
        mean_matched_score = _safe_divide(sum_matched_score, tp)
        # the score average over all gt/true objects
        mean_true_score = _safe_divide(sum_matched_score, n_true)
        panoptic_quality = _safe_divide(sum_matched_score, tp + fp / 2 + fn / 2)

        stats_dict = dict(
            criterion=criterion,
            thresh=thr,
            fp=fp,
            tp=tp,
            fn=fn,
            precision=precision(tp, fp, fn),
            recall=recall(tp, fp, fn),
            accuracy=accuracy(tp, fp, fn),
            f1=f1(tp, fp, fn),
            n_true=n_true,
            n_pred=n_pred,
            mean_true_score=mean_true_score,
            mean_matched_score=mean_matched_score,
            panoptic_quality=panoptic_quality,
        )
        if bool(report_matches):
            if not_trivial:
                stats_dict.update(
                    # int() to be json serializable
                    matched_pairs=tuple(
                        (int(map_rev_true[i]), int(map_rev_pred[j]))
                        for i, j in zip(1 + true_ind, 1 + pred_ind)
                    ),
                    matched_scores=tuple(scores[true_ind, pred_ind]),
                    matched_tps=tuple(map(int, np.flatnonzero(match_ok))),
                )
            else:
                stats_dict.update(
                    matched_pairs=(),
                    matched_scores=(),
                    matched_tps=(),
                )
        return namedtuple("Matching", stats_dict.keys())(*stats_dict.values())

    return _single(thresh) if np.isscalar(thresh) else tuple(map(_single, thresh))


def matching_dataset(
    y_true,
    y_pred,
    thresh=0.5,
    criterion="iou",
    by_image=False,
    show_progress=True,
    parallel=False,
):
    """matching metrics for list of images, see `stardist.matching.matching`"""
    len(y_true) == len(y_pred) or ValueError(
        "y_true and y_pred must have the same length."
    )
    return matching_dataset_lazy(
        tuple(zip(y_true, y_pred)),
        thresh=thresh,
        criterion=criterion,
        by_image=by_image,
        show_progress=show_progress,
        parallel=parallel,
    )


def matching_dataset_lazy(
    y_gen,
    thresh=0.5,
    criterion="iou",
    by_image=False,
    show_progress=True,
    parallel=False,
):
    expected_keys = set(
        (
            "fp",
            "tp",
            "fn",
            "precision",
            "recall",
            "accuracy",
            "f1",
            "criterion",
            "thresh",
            "n_true",
            "n_pred",
            "mean_true_score",
            "mean_matched_score",
            "panoptic_quality",
        )
    )

    single_thresh = False
    if np.isscalar(thresh):
        single_thresh = True
        thresh = (thresh,)

    tqdm_kwargs = {}
    tqdm_kwargs["disable"] = not bool(show_progress)
    if int(show_progress) > 1:
        tqdm_kwargs["total"] = int(show_progress)

    # compute matching stats for every pair of label images
    if parallel:
        from concurrent.futures import ThreadPoolExecutor

        fn = lambda pair: matching(
            *pair, thresh=thresh, criterion=criterion, report_matches=False
        )
        with ThreadPoolExecutor() as pool:
            stats_all = tuple(pool.map(fn, tqdm(y_gen, **tqdm_kwargs)))
    else:
        stats_all = tuple(
            matching(y_t, y_p, thresh=thresh, criterion=criterion, report_matches=False)
            for y_t, y_p in tqdm(y_gen, **tqdm_kwargs)
        )

    # accumulate results over all images for each threshold separately
    n_images, n_threshs = len(stats_all), len(thresh)
    accumulate = [{} for _ in range(n_threshs)]
    for stats in stats_all:
        for i, s in enumerate(stats):
            acc = accumulate[i]
            for k, v in s._asdict().items():
                if k == "mean_true_score" and not bool(by_image):
                    # convert mean_true_score to "sum_matched_score"
                    acc[k] = acc.setdefault(k, 0) + v * s.n_true
                else:
                    try:
                        acc[k] = acc.setdefault(k, 0) + v
                    except TypeError:
                        pass

    # normalize/compute 'precision', 'recall', 'accuracy', 'f1'
    for thr, acc in zip(thresh, accumulate):
        set(acc.keys()) == expected_keys or ValueError("unexpected keys")
        acc["criterion"] = criterion
        acc["thresh"] = thr
        acc["by_image"] = bool(by_image)
        if bool(by_image):
            for k in (
                "precision",
                "recall",
                "accuracy",
                "f1",
                "mean_true_score",
                "mean_matched_score",
                "panoptic_quality",
            ):
                acc[k] /= n_images
        else:
            tp, fp, fn, n_true = acc["tp"], acc["fp"], acc["fn"], acc["n_true"]
            sum_matched_score = acc["mean_true_score"]

            mean_matched_score = _safe_divide(sum_matched_score, tp)
            mean_true_score = _safe_divide(sum_matched_score, n_true)
            panoptic_quality = _safe_divide(sum_matched_score, tp + fp / 2 + fn / 2)

            acc.update(
                precision=precision(tp, fp, fn),
                recall=recall(tp, fp, fn),
                accuracy=accuracy(tp, fp, fn),
                f1=f1(tp, fp, fn),
                mean_true_score=mean_true_score,
                mean_matched_score=mean_matched_score,
                panoptic_quality=panoptic_quality,
            )

    accumulate = tuple(
        namedtuple("DatasetMatching", acc.keys())(*acc.values()) for acc in accumulate
    )
    return accumulate[0] if single_thresh else accumulate


# copied from scikit-image master for now (remove when part of a release)
def relabel_sequential(label_field, offset=1):
    """Relabel arbitrary labels to {`offset`, ... `offset` + number_of_labels}.

    This function also returns the forward map (mapping the original labels to
    the reduced labels) and the inverse map (mapping the reduced labels back
    to the original ones).

    Parameters
    ----------
    label_field : numpy array of int, arbitrary shape
        An array of labels, which must be non-negative integers.
    offset : int, optional
        The return labels will start at `offset`, which should be
        strictly positive.

    Returns
    -------
    relabeled : numpy array of int, same shape as `label_field`
        The input label field with labels mapped to
        {offset, ..., number_of_labels + offset - 1}.
        The data type will be the same as `label_field`, except when
        offset + number_of_labels causes overflow of the current data type.
    forward_map : numpy array of int, shape ``(label_field.max() + 1,)``
        The map from the original label space to the returned label
        space. Can be used to re-apply the same mapping. See examples
        for usage. The data type will be the same as `relabeled`.
    inverse_map : 1D numpy array of int, of length offset + number of labels
        The map from the new label space to the original space. This
        can be used to reconstruct the original label field from the
        relabeled one. The data type will be the same as `relabeled`.

    Notes
    -----
    The label 0 is assumed to denote the background and is never remapped.

    The forward map can be extremely big for some inputs, since its
    length is given by the maximum of the label field. However, in most
    situations, ``label_field.max()`` is much smaller than
    ``label_field.size``, and in these cases the forward map is
    guaranteed to be smaller than either the input or output images.

    Examples
    --------
    >>> from skimage.segmentation import relabel_sequential
    >>> label_field = np.array([1, 1, 5, 5, 8, 99, 42])
    >>> relab, fw, inv = relabel_sequential(label_field)
    >>> relab
    array([1, 1, 2, 2, 3, 5, 4])
    >>> fw
    array([0, 1, 0, 0, 0, 2, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5])
    >>> inv
    array([ 0,  1,  5,  8, 42, 99])
    >>> (fw[label_field] == relab).all()
    True
    >>> (inv[relab] == label_field).all()
    True
    >>> relab, fw, inv = relabel_sequential(label_field, offset=5)
    >>> relab
    array([5, 5, 6, 6, 7, 9, 8])
    """
    offset = int(offset)
    if offset <= 0:
        raise ValueError("Offset must be strictly positive.")
    if np.min(label_field) < 0:
        raise ValueError("Cannot relabel array that contains negative values.")
    max_label = int(label_field.max())  # Ensure max_label is an integer
    if not np.issubdtype(label_field.dtype, np.integer):
        new_type = np.min_scalar_type(max_label)
        label_field = label_field.astype(new_type)
    labels = np.unique(label_field)
    labels0 = labels[labels != 0]
    new_max_label = offset - 1 + len(labels0)
    new_labels0 = np.arange(offset, new_max_label + 1)
    output_type = label_field.dtype
    required_type = np.min_scalar_type(new_max_label)
    if np.dtype(required_type).itemsize > np.dtype(label_field.dtype).itemsize:
        output_type = required_type
    forward_map = np.zeros(max_label + 1, dtype=output_type)
    forward_map[labels0] = new_labels0
    inverse_map = np.zeros(new_max_label + 1, dtype=output_type)
    inverse_map[offset:] = labels0
    relabeled = forward_map[label_field]
    return relabeled, forward_map, inverse_map


def group_matching_labels(ys, thresh=1e-10, criterion="iou"):
    """
    Group matching objects (i.e. assign the same label id) in a
    list of label images (e.g. consecutive frames of a time-lapse).

    Uses function `matching` (with provided `criterion` and `thresh`) to
    iteratively/greedily match and group objects/labels in consecutive images of `ys`.
    To that end, matching objects are grouped together by assigning the same label id,
    whereas unmatched objects are assigned a new label id.
    At the end of this process, each label group will have been assigned a unique id.

    Note that the label images `ys` will not be modified. Instead, they will initially
    be duplicated and converted to data type `np.int32` before objects are grouped and the result
    is returned. (Note that `np.int32` limits the number of label groups to at most 2147483647.)

    Example
    -------
    import numpy as np
    from stardist.data import test_image_nuclei_2d
    from stardist.matching import group_matching_labels

    _y = test_image_nuclei_2d(return_mask=True)[1]
    labels = np.stack([_y, 2*np.roll(_y,10)], axis=0)

    labels_new = group_matching_labels(labels)

    Parameters
    ----------
    ys : np.ndarray or list/tuple of np.ndarray
        list/array of integer labels (2D or 3D)

    """
    # check 'ys' without making a copy
    len(ys) > 1 or ValueError("'ys' must have 2 or more entries")
    if isinstance(ys, np.ndarray):
        _check_label_array(ys, "ys")
        ys.ndim > 1 or ValueError("'ys' must be at least 2-dimensional")
        ys_grouped = np.empty_like(ys, dtype=np.int32)
    else:
        all(_check_label_array(y, "ys") for y in ys) or ValueError(
            "'ys' must be a list of label images"
        )
        all(y.shape == ys[0].shape for y in ys) or ValueError(
            "all label images must have the same shape"
        )
        ys_grouped = np.empty((len(ys),) + ys[0].shape, dtype=np.int32)

    def _match_single(y_prev, y, next_id):
        y = y.astype(np.int32, copy=False)
        res = matching(
            y_prev, y, report_matches=True, thresh=thresh, criterion=criterion
        )
        # relabel dict (for matching labels) that maps label ids from y -> y_prev
        relabel = dict(reversed(res.matched_pairs[i]) for i in res.matched_tps)
        y_grouped = np.zeros_like(y)
        for r in regionprops(y):
            m = y[r.slice] == r.label
            if r.label in relabel:
                y_grouped[r.slice][m] = relabel[r.label]
            else:
                y_grouped[r.slice][m] = next_id
                next_id += 1
        return y_grouped, next_id

    ys_grouped[0] = ys[0]
    next_id = ys_grouped[0].max() + 1
    for i in range(len(ys) - 1):
        ys_grouped[i + 1], next_id = _match_single(ys_grouped[i], ys[i + 1], next_id)
    return ys_grouped


def _shuffle_labels(y):
    _check_label_array(y, "y")
    y2 = np.zeros_like(y)
    ids = tuple(set(np.unique(y)) - {0})
    relabel = dict(zip(ids, np.random.permutation(ids)))
    for r in regionprops(y):
        m = y[r.slice] == r.label
        y2[r.slice][m] = relabel[r.label]
    return y2



from typing import Union

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F

import torchist
from torch import optim, nn as nn

from pytorch_rdc_net.matching import matching
from pytorch_rdc_net.stacked_dilated_conv import StackedDilatedConv
from pytorch_rdc_net.utils import get_dimensional_modules


class RDCNet(pl.LightningModule):
    def __init__(
        self,
        dim: int,
        in_channels: int = 1,
        down_sampling_factor: tuple = (6, 6),
        down_sampling_channels: int = 8,
        spatial_dropout_p: float = 0.1,
        channels_per_group: int = 32,
        n_groups: int = 4,
        dilation_rates: list[int] = [1, 2, 4, 8, 16],
        steps: int = 6,
        lr: float = 0.001,
        instance_size: tuple = (20.0, 20.0),
    ):
        super(RDCNet, self).__init__()
        self.save_hyperparameters()
        self.dim = dim

        (
            Conv,
            ConvTranspose,
            Dropout,
            lovasz_softmax,
            InstanceEmbeddingLoss,
            self.median_filter,
        ) = get_dimensional_modules(dim)

        down_sampling_kernel = tuple(
            [max(3, dsf if dsf % 2 != 0 else dsf + 1) for dsf in down_sampling_factor]
        )
        self.in_conv = Conv(
            in_channels=self.hparams.in_channels,
            out_channels=self.hparams.down_sampling_channels,
            kernel_size=down_sampling_kernel,
            stride=self.hparams.down_sampling_factor,
            padding=tuple([k // 2 for k in down_sampling_kernel]),
        )

        self.spatial_dropout = Dropout(p=self.hparams.spatial_dropout_p)

        self.reduce_ch_conv = Conv(
            in_channels=self.hparams.channels_per_group * self.hparams.n_groups
            + self.hparams.down_sampling_channels,
            out_channels=self.hparams.channels_per_group * self.hparams.n_groups,
            kernel_size=1,
        )

        self.sd_conv = StackedDilatedConv(
            in_channels=self.hparams.channels_per_group * self.hparams.n_groups,
            out_channels=self.hparams.channels_per_group * self.hparams.n_groups,
            kernel_size=3,
            dilation_rates=self.hparams.dilation_rates,
            groups=self.hparams.n_groups,
            dim=self.dim,
        )

        up_sampling_kernel = tuple(
            [2 * dsk if dsk > 3 else 3 for dsk in down_sampling_kernel]
        )
        up_padding = tuple(
            [
                usk // 2 if dsf <= 3 else dsf // 2 + 1
                for dsf, usk in zip(down_sampling_factor, down_sampling_kernel)
            ]
        )
        output_padding = tuple(
            [
                -(-1 * s - 2 * p + (k - 1) + 1) % s
                for k, p, s in zip(
                    up_sampling_kernel, up_padding, self.hparams.down_sampling_factor
                )
            ]
        )
        self.out_conv = ConvTranspose(
            in_channels=self.hparams.channels_per_group * self.hparams.n_groups,
            out_channels=7 if dim == 2 else 11,
            kernel_size=up_sampling_kernel,
            stride=self.hparams.down_sampling_factor,
            padding=up_padding,
            output_padding=output_padding,
        )

        self.embedding_loss = InstanceEmbeddingLoss()
        self.semantic_loss = lovasz_softmax

        self.receptive_field_size = None
        self.coords = None

    def forward(self, x):
        assert all(
            x.shape[i + 2] % self.hparams.down_sampling_factor[i] == 0
            for i in range(self.dim)
        )
        x = self.in_conv(x)

        state = torch.zeros(
            x.shape[0],
            self.hparams.channels_per_group * self.hparams.n_groups,
            *x.shape[2:],
            dtype=x.dtype,
            device=x.device,
        )

        for _ in range(self.hparams.steps):
            delta = torch.cat([x, state], dim=1)
            delta = self.spatial_dropout(delta)
            delta = F.leaky_relu(delta)
            delta = self.reduce_ch_conv(delta)
            delta = F.leaky_relu(delta)
            delta = self.sd_conv(delta)
            state += delta

        state = F.leaky_relu(state)
        state = self.out_conv(state)
        embeddings = state[:, : self.dim]
        lower_triangular = state[
            :, self.dim : self.dim + (self.dim * (self.dim + 1)) // 2
        ]
        semantic_classes = F.softmax(
            state[:, self.dim + (self.dim * (self.dim + 1)) // 2 :], dim=1
        )

        return embeddings, lower_triangular, semantic_classes

    def predict_instances(self, x, vote_masking_threshold=0.01):
        """
        Predict instance segmentations for the input batch.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, channels, *spatial_dims).
            vote_masking_threshold (float): Threshold for vote masking.

        Returns:
            np.ndarray: Instance segmentations for the input batch.
        """
        self.eval()
        instance_segmentations = []
        with torch.no_grad():
            for patch in x:
                # Forward pass
                embeddings, lower_triangular, semantic = self(
                    patch.unsqueeze(0).to(self.device)
                )

                # Get instance segmentations
                label_img = self.get_instance_segmentations(
                    embeddings[0],
                    lower_triangular[0],
                    semantic[0],
                    vote_masking_threshold=vote_masking_threshold,
                )
                instance_segmentations.append(label_img.cpu().numpy()[np.newaxis])

        return np.stack(instance_segmentations)

    def get_instance_segmentations(
        self,
        embeddings,
        lower_triangular,
        semantic,
        vote_masking_threshold=0.01,
        max_instances: int = None,
    ):
        """
        Generate instance segmentations from embeddings, lower triangular matrix, and semantic predictions.

        Args:
            embeddings (torch.Tensor): Embedding tensor of shape (*spatial_dims, dim).
            lower_triangular (torch.Tensor): Lower triangular matrix tensor.
            semantic (torch.Tensor): Semantic predictions tensor.
            vote_masking_threshold (float): Threshold for vote masking.
            max_instances (int, optional): Maximum number of instances to predict. Defaults to None, which computes all instances.

        Returns:
            torch.Tensor: Instance segmentation map.
        """
        if max_instances is None:
            max_instances = torch.inf
        with torch.no_grad():
            embeddings = embeddings.detach()
            lower_triangular = lower_triangular.detach()
            semantic = semantic.detach()

            embeddings, lower_triangular, padding, semantic = self.pad_inputs(
                embeddings, lower_triangular, semantic
            )

            shape = embeddings.shape[-self.dim :]

            fg_mask = torch.argmax(semantic, dim=0).type(torch.bool)

            # Add coordinate grid to embeddings
            grid = self._get_coordinate_grid(embeddings)[0]
            embeddings = embeddings + grid

            # Histogram-based voting
            fg_embeddings = embeddings[:, fg_mask]
            bins = tuple(shape[i] for i in range(self.dim))
            low = tuple(0 for _ in range(self.dim))
            upp = tuple(
                shape[i] / self.hparams.instance_size[i] for i in range(self.dim)
            )
            votes = torchist.histogramdd(
                fg_embeddings.moveaxis(0, -1), bins=bins, low=low, upp=upp
            )
            votes[fg_mask == 0] = 0

            label_img = torch.zeros_like(votes, dtype=torch.int32, device=votes.device)
            label_id = 0

            while votes.max() > 0 and label_id < max_instances:
                center = torch.unravel_index(torch.argmax(votes), votes.shape)
                cov = self.construct_cov_matrix(center, lower_triangular)
                instance = self.compute_instance_probabilities(
                    center, cov, embeddings, grid, shape
                )

                instance[label_img > 0] = 0

                if torch.sum(instance) > 0:
                    label_id += 1
                    label_img[self.median_filter(instance) >= 0.5] = label_id

                fg_mask[*center] = 0
                fg_mask[instance >= vote_masking_threshold] = 0
                fg_embeddings = embeddings[:, fg_mask]

                votes = torchist.histogramdd(
                    fg_embeddings.moveaxis(0, -1), bins=bins, low=low, upp=upp
                )
                votes[fg_mask == 0] = 0

            # Remove padding
            slices = tuple(
                slice(padding[i * 2], -padding[i * 2 + 1])
                for i in reversed(range(self.dim))
            )
            return label_img[slices].detach().cpu()

    def pad_inputs(self, embeddings, lower_triangular, semantic):
        padding = [int(self.hparams.instance_size[i]) + 1 for i in range(self.dim)]
        padding = tuple(reversed([p for p in padding for _ in range(2)]))
        embeddings = F.pad(embeddings, padding)
        lower_triangular = F.pad(lower_triangular, padding)
        semantic = F.pad(semantic, padding)
        return embeddings, lower_triangular, padding, semantic

    def compute_instance_probabilities(self, center, cov, embeddings, grid, shape):
        target_centroid = grid[:, *center]
        for _ in range(self.dim):
            target_centroid = target_centroid.unsqueeze(-1)
        c_e = torch.square(embeddings - target_centroid)
        for _ in range(self.dim - 2):
            c_e = c_e.unsqueeze(-1)
        c_e = torch.moveaxis(c_e.view(self.dim, 1, -1), -1, 0)
        instance = torch.exp(-1 * (torch.transpose(c_e, 1, 2) @ cov @ c_e).view(*shape))
        return instance

    def construct_cov_matrix(self, center, lower_triangular):
        lt_matrix = torch.zeros(
            (self.dim, self.dim),
            device=lower_triangular.device,
            dtype=lower_triangular.dtype,
        )
        idx = 0
        for i in range(self.dim):
            lt_matrix[i, i] = F.softplus(lower_triangular[idx, *center])
            idx += 1
            for j in range(i + 1, self.dim):
                lt_matrix[j, i] = lower_triangular[idx, *center]
                idx += 1
        cov = lt_matrix @ lt_matrix.transpose(0, 1)
        return cov

    def _get_coordinate_grid(self, pred: torch.Tensor) -> torch.Tensor:
        if (
            self.coords is None
            or self.coords.shape[-self.dim :] != pred.shape[-self.dim :]
        ):
            grid = torch.meshgrid(
                [torch.arange(size) for size in pred.shape[-self.dim :]],
                indexing="ij",
            )
            grid = torch.stack(grid, dim=0).type(torch.float32)
            for i in range(self.dim):
                grid[i] = grid[i] / self.hparams.instance_size[i]
            grid = grid.unsqueeze(0)

            grid.requires_grad = False
            self.coords = grid.to(pred.device)

        return self.coords

    def training_step(self, batch, batch_idx):
        """
        Perform a single training step.

        Args:
            batch (tuple): A tuple containing input data and ground truth labels.
            batch_idx (int): Index of the batch.

        Returns:
            torch.Tensor: The total training loss for the batch.
        """
        x, gt_labels = batch
        embeddings, lower_triangular, semantic_classes = self(x)
        grid = self._get_coordinate_grid(embeddings)

        # Compute embedding loss
        embedding_loss, var_lt = self.embedding_loss(
            grid[0], embeddings + grid, lower_triangular, gt_labels
        )

        # Compute semantic loss
        semantic_loss = self.semantic_loss(
            semantic_classes, gt_labels > 0, per_image=True
        )

        # Total training loss
        train_loss = embedding_loss + semantic_loss + var_lt

        # Log metrics
        self.log(
            "semantic_loss",
            semantic_loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "embedding_loss",
            embedding_loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "var_lt", var_lt, on_step=False, on_epoch=True, prog_bar=True, logger=True
        )
        self.log(
            "train_loss",
            train_loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return train_loss

    def validation_step(self, batch, batch_idx):
        """
        Perform a single validation step.

        Args:
            batch (tuple): A tuple containing input data and ground truth labels.
            batch_idx (int): Index of the batch.

        Returns:
            float: Mean true score for the batch.
        """
        x, gt_labels = batch
        gt = gt_labels.cpu().numpy()[0, 0]

        embeddings, lower_triangular, semantic_classes = self(x)

        # Generate instance segmentations
        instance_seg = self.get_instance_segmentations(
            embeddings[0],
            lower_triangular[0],
            semantic_classes[0],
            vote_masking_threshold=0.01,
            max_instances=2 * torch.max(gt_labels).item(),
        )

        # Compute metrics
        metrics = matching(
            y_true=gt,
            y_pred=instance_seg.cpu().numpy(),
            criterion="iou",
        )

        # Log metrics
        self.log(
            "precision",
            metrics.precision,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "recall",
            metrics.recall,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "f1", metrics.f1, on_step=False, on_epoch=True, prog_bar=True, logger=True
        )
        self.log(
            "mean_matched_score",
            metrics.mean_matched_score,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "mean_true_score",
            metrics.mean_true_score,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "panoptic_quality",
            metrics.panoptic_quality,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return metrics.mean_true_score

    def test_step(self, batch, batch_idx):
        """
        Perform a single test step.

        Args:
            batch (tuple): A tuple containing input data and ground truth labels.
            batch_idx (int): Index of the batch.

        Returns:
            float: Mean true score for the batch.
        """
        x, gt_labels = batch
        gt = gt_labels.cpu().numpy()[0, 0]

        # Forward pass
        embeddings, lower_triangular, semantic_classes = self(x)

        # Generate instance segmentations
        instance_seg = self.get_instance_segmentations(
            embeddings[0],
            lower_triangular[0],
            semantic_classes[0],
            vote_masking_threshold=0.01,
        )

        # Compute metrics
        metrics = matching(
            y_true=gt,
            y_pred=instance_seg.cpu().numpy(),
            criterion="iou",
        )

        # Log metrics
        self.log(
            "test_precision",
            metrics.precision,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "test_recall",
            metrics.recall,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "test_f1",
            metrics.f1,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "test_mean_matched_score",
            metrics.mean_matched_score,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "test_mean_true_score",
            metrics.mean_true_score,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )
        self.log(
            "test_panoptic_quality",
            metrics.panoptic_quality,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        return metrics.mean_true_score

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr)
        scheduler = optim.lr_scheduler.CosineAnnealingLR(
            optimizer=optimizer, T_max=self.trainer.max_epochs, eta_min=1e-5
        )
        return {"optimizer": optimizer, "lr_scheduler": scheduler}

    def get_receptive_field(self):
        """
        Calculate the receptive field of the network.

        Returns:
            tuple: Receptive field and its dimensions.
        """
        if self.receptive_field_size is None:
            rf = compute_receptive_field(
                dim=self.hparams.dim,
                down_sampling_factor=self.hparams.down_sampling_factor,
                dilation_rates=self.hparams.dilation_rates,
                steps=self.hparams.steps,
            )
            if self.dim == 2:
                h = np.where(rf.max((0, 2)))
                w = np.where(rf.max((0, 1)))
                self.receptive_field_size = (
                    max(h[0][-1] - h[0][0], w[0][-1] - w[0][0]),
                ) * 2
            elif self.dim == 3:
                d = np.where(rf.max((0, 2, 3)))
                h = np.where(rf.max((0, 1, 3)))
                w = np.where(rf.max((0, 1, 2)))
                self.receptive_field_size = (d[0][-1] - d[0][0],) + (
                    max(h[0][-1] - h[0][0], w[0][-1] - w[0][0]),
                ) * 2

        return self.receptive_field_size


def compute_receptive_field(
    dim: int,
    down_sampling_factor: Union[tuple[int, int], tuple[int, int, int]],
    dilation_rates: list[int],
    steps: int,
):
    rf_net = RDCNet(
        dim=dim,
        in_channels=1,
        down_sampling_factor=down_sampling_factor,
        down_sampling_channels=1,
        spatial_dropout_p=0,
        channels_per_group=1,
        n_groups=1,
        dilation_rates=dilation_rates,
        steps=steps,
    )
    for param in rf_net.parameters():
        param.requires_grad = False
        param.data = nn.parameter.Parameter(torch.ones_like(param))

    rf_net.eval()
    with torch.no_grad():
        shape = tuple(
            [4 * dsf * max(dilation_rates) * steps for dsf in down_sampling_factor]
        )
        # Create input tensor with a single activated point
        x = torch.zeros((1, 1, *shape), dtype=torch.float32)
        center = tuple(s // 2 for s in shape)
        x[(0, 0, *center)] = 1
        p_0 = rf_net(x)[0].detach()
        x[(0, 0, *center)] = 10
        p_1 = rf_net(x)[0].detach()

    # Calculate receptive field
    return (torch.abs(p_1 - p_0) > 0).sum(0).detach().cpu().numpy()



from torch import nn as nn
from torch.nn import functional as F

from pytorch_rdc_net.embedding_loss import (
    InstanceEmbeddingLoss2d,
    InstanceEmbeddingLoss3d,
)
from pytorch_rdc_net.lovasz_losses import lovasz_softmax_2d, lovasz_softmax_3d


def get_dimensional_modules(dim: int):
    """
    Get the appropriate dimensional modules based on the specified dimension.

    Parameters:
        dim (int): The dimension (2 or 3) for which to get the modules.

    Returns:
        tuple: A tuple containing the appropriate Conv, ConvTranspose, Dropout,
        lovasz_softmax, and InstanceEmbeddingLoss classes.
    """
    if dim == 2:
        return (
            nn.Conv2d,
            nn.ConvTranspose2d,
            nn.Dropout2d,
            lovasz_softmax_2d,
            InstanceEmbeddingLoss2d,
            median_filter_2d,
        )
    elif dim == 3:
        return (
            nn.Conv3d,
            nn.ConvTranspose3d,
            nn.Dropout3d,
            lovasz_softmax_3d,
            InstanceEmbeddingLoss3d,
            median_filter_3d,
        )
    else:
        raise ValueError("Invalid dimension. Only 2D and 3D are supported.")


def median_filter_2d(input_tensor):
    patches = (
        F.pad(input_tensor, (1, 1, 1, 1), mode="constant")
        .unfold(0, 3, 1)
        .unfold(1, 3, 1)
        .contiguous()
        .view(-1, 9)
    )
    return patches.median(dim=1)[0].view(input_tensor.shape)


def median_filter_3d(input_tensor):
    patches = (
        F.pad(input_tensor, (1, 1, 1, 1, 1, 1), mode="constant")
        .unfold(0, 3, 1)
        .unfold(1, 3, 1)
        .unfold(2, 3, 1)
        .contiguous()
        .view(-1, 27)
    )
    return patches.median(dim=1)[0].view(input_tensor.shape)



"""
MIT License

Copyright (c) 2018 Maxim Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Lovasz-Softmax and Jaccard hinge loss in PyTorch
Maxim Berman 2018 ESAT-PSI KU Leuven (MIT License)

----------
Extended by:
- Tim-Oliver Buchholz, Friedrich Miescher Institute for Biomedical Research, 2025
    * Added 3D support.
"""

from __future__ import print_function, division

import torch
from torch.autograd import Variable
import torch.nn.functional as F
import numpy as np

try:
    from itertools import ifilterfalse
except ImportError:  # py3k
    from itertools import filterfalse as ifilterfalse


def lovasz_grad(gt_sorted):
    """
    Computes gradient of the Lovasz extension w.r.t sorted errors
    See Alg. 1 in paper
    """
    p = len(gt_sorted)
    gts = gt_sorted.sum()
    intersection = gts - gt_sorted.float().cumsum(0)
    union = gts + (1 - gt_sorted).float().cumsum(0)
    jaccard = 1.0 - intersection / union
    if p > 1:  # cover 1-pixel case
        jaccard[1:p] = jaccard[1:p] - jaccard[0:-1]
    return jaccard


def iou_binary(preds, labels, EMPTY=1.0, ignore=None, per_image=True):
    """
    IoU for foreground class
    binary: 1 foreground, 0 background
    """
    if not per_image:
        preds, labels = (preds,), (labels,)
    ious = []
    for pred, label in zip(preds, labels):
        intersection = ((label == 1) & (pred == 1)).sum()
        union = ((label == 1) | ((pred == 1) & (label != ignore))).sum()
        if not union:
            iou = EMPTY
        else:
            iou = float(intersection) / float(union)
        ious.append(iou)
    iou = mean(ious)  # mean across images if per_image
    return 100 * iou


def iou(preds, labels, C, EMPTY=1.0, ignore=None, per_image=False):
    """
    Array of IoU for each (non ignored) class
    """
    if not per_image:
        preds, labels = (preds,), (labels,)
    ious = []
    for pred, label in zip(preds, labels):
        iou = []
        for i in range(C):
            if (
                i != ignore
            ):  # The ignored label is sometimes among predicted classes (ENet - CityScapes)
                intersection = ((label == i) & (pred == i)).sum()
                union = ((label == i) | ((pred == i) & (label != ignore))).sum()
                if not union:
                    iou.append(EMPTY)
                else:
                    iou.append(float(intersection) / float(union))
        ious.append(iou)
    ious = [mean(iou) for iou in zip(*ious)]  # mean across images if per_image
    return 100 * np.array(ious)


# --------------------------- BINARY LOSSES ---------------------------


def lovasz_hinge(logits, labels, per_image=True, ignore=None):
    """
    Binary Lovasz hinge loss
      logits: [B, H, W] Variable, logits at each pixel (between -\infty and +\infty)
      labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
      per_image: compute the loss per image instead of per batch
      ignore: void class id
    """
    if per_image:
        loss = mean(
            lovasz_hinge_flat(
                *flatten_binary_scores(log.unsqueeze(0), lab.unsqueeze(0), ignore)
            )
            for log, lab in zip(logits, labels)
        )
    else:
        loss = lovasz_hinge_flat(*flatten_binary_scores(logits, labels, ignore))
    return loss


def lovasz_hinge3d(logits, labels, per_image=True, ignore=None):
    """
    Binary Lovasz hinge loss
      logits: [B, D, H, W] Variable, logits at each pixel (between -\infty and +\infty)
      labels: [B, D, H, W] Tensor, binary ground truth masks (0 or 1)
      per_image: compute the loss per image instead of per batch
      ignore: void class id
    """
    if per_image:
        loss = mean(
            lovasz_hinge_flat3d(
                *flatten_binary_scores3d(log.unsqueeze(0), lab.unsqueeze(0), ignore)
            )
            for log, lab in zip(logits, labels)
        )
    else:
        loss = lovasz_hinge_flat3d(*flatten_binary_scores3d(logits, labels, ignore))
    return loss


def lovasz_hinge_flat(logits, labels):
    """
    Binary Lovasz hinge loss
      logits: [P] Variable, logits at each prediction (between -\infty and +\infty)
      labels: [P] Tensor, binary ground truth labels (0 or 1)
      ignore: label to ignore
    """
    if len(labels) == 0:
        # only void pixels, the gradients should be 0
        return logits.sum() * 0.0
    signs = 2.0 * labels.float() - 1.0
    errors = 1.0 - logits * Variable(signs)
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), Variable(grad))
    return loss


def lovasz_hinge_flat3d(logits, labels):
    """
    Binary Lovasz hinge loss
      logits: [P] Variable, logits at each prediction (between -\infty and +\infty)
      labels: [P] Tensor, binary ground truth labels (0 or 1)
      ignore: label to ignore
    """
    if len(labels) == 0:
        # only void pixels, the gradients should be 0
        return logits.sum() * 0.0
    signs = 2.0 * labels.float() - 1.0
    errors = 1.0 - logits * Variable(signs)
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), Variable(grad))
    return loss


def flatten_binary_scores(scores, labels, ignore=None):
    """
    Flattens predictions in the batch (binary case)
    Remove labels equal to 'ignore'
    """
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = labels != ignore
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels


def flatten_binary_scores3d(scores, labels, ignore=None):
    """
    Flattens predictions in the batch (binary case)
    Remove labels equal to 'ignore'
    """
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = labels != ignore
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels


class StableBCELoss(torch.nn.modules.Module):
    def __init__(self):
        super(StableBCELoss, self).__init__()

    def forward(self, input, target):
        neg_abs = -input.abs()
        loss = input.clamp(min=0) - input * target + (1 + neg_abs.exp()).log()
        return loss.mean()


def binary_xloss(logits, labels, ignore=None):
    """
    Binary Cross entropy loss
      logits: [B, H, W] Variable, logits at each pixel (between -\infty and +\infty)
      labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
      ignore: void class id
    """
    logits, labels = flatten_binary_scores(logits, labels, ignore)
    loss = StableBCELoss()(logits, Variable(labels.float()))
    return loss


# --------------------------- MULTICLASS LOSSES ---------------------------


def lovasz_softmax_2d(probas, labels, classes="present", per_image=False, ignore=None):
    """
    Multi-class Lovasz-Softmax loss
      probas: [B, C, H, W] Variable, class probabilities at each prediction (between 0 and 1).
              Interpreted as binary (sigmoid) output with outputs of size [B, H, W].
      labels: [B, H, W] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
      per_image: compute the loss per image instead of per batch
      ignore: void class labels
    """
    if per_image:
        loss = mean(
            lovasz_softmax_flat(
                *flatten_probas(prob.unsqueeze(0), lab.unsqueeze(0), ignore),
                classes=classes,
            )
            for prob, lab in zip(probas, labels)
        )
    else:
        loss = lovasz_softmax_flat(
            *flatten_probas(probas, labels, ignore), classes=classes
        )
    return loss


def lovasz_softmax_3d(probas, labels, classes="present", per_image=False, ignore=None):
    """
    Multi-class Lovasz-Softmax loss
      probas: [B, C, D, H, W] Variable, class probabilities at each prediction (between 0 and 1).
              Interpreted as binary (sigmoid) output with outputs of size [B, H, W].
      labels: [B, D, H, W] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
      per_image: compute the loss per image instead of per batch
      ignore: void class labels
    """
    if per_image:
        loss = mean(
            lovasz_softmax_flat3d(
                *flatten_probas3d(prob.unsqueeze(0), lab.unsqueeze(0), ignore),
                classes=classes,
            )
            for prob, lab in zip(probas, labels)
        )
    else:
        loss = lovasz_softmax_flat3d(
            *flatten_probas3d(probas, labels, ignore), classes=classes
        )
    return loss


def lovasz_softmax_flat(probas, labels, classes="present"):
    """
    Multi-class Lovasz-Softmax loss
      probas: [P, C] Variable, class probabilities at each prediction (between 0 and 1)
      labels: [P] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
    """
    if probas.numel() == 0:
        # only void pixels, the gradients should be 0
        return probas * 0.0
    C = probas.size(1)
    losses = []
    class_to_sum = list(range(C)) if classes in ["all", "present"] else classes
    for c in class_to_sum:
        fg = (labels == c).float()  # foreground for class c
        if classes == "present" and fg.sum() == 0:
            continue
        if C == 1:
            if len(classes) > 1:
                raise ValueError("Sigmoid output possible only with 1 class")
            class_pred = probas[:, 0]
        else:
            class_pred = probas[:, c]
        errors = (Variable(fg) - class_pred).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, Variable(lovasz_grad(fg_sorted))))
    return mean(losses)


def lovasz_softmax_flat3d(probas, labels, classes="present"):
    """
    Multi-class Lovasz-Softmax loss
      probas: [P, C] Variable, class probabilities at each prediction (between 0 and 1)
      labels: [P] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
    """
    if probas.numel() == 0:
        # only void pixels, the gradients should be 0
        return probas * 0.0
    C = probas.size(1)
    losses = []
    class_to_sum = list(range(C)) if classes in ["all", "present"] else classes
    for c in class_to_sum:
        fg = (labels == c).float()  # foreground for class c
        if classes == "present" and fg.sum() == 0:
            continue
        if C == 1:
            if len(classes) > 1:
                raise ValueError("Sigmoid output possible only with 1 class")
            class_pred = probas[:, 0]
        else:
            class_pred = probas[:, c]
        errors = (Variable(fg) - class_pred).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, Variable(lovasz_grad(fg_sorted))))
    return mean(losses)


def flatten_probas(probas, labels, ignore=None):
    """
    Flattens predictions in the batch
    """
    if probas.dim() == 3:
        # assumes output of a sigmoid layer
        B, H, W = probas.size()
        probas = probas.view(B, 1, H, W)
    B, C, H, W = probas.size()
    probas = probas.permute(0, 2, 3, 1).contiguous().view(-1, C)  # B * H * W, C = P, C
    labels = labels.view(-1)
    if ignore is None:
        return probas, labels
    valid = labels != ignore
    vprobas = probas[valid.nonzero().squeeze()]
    vlabels = labels[valid]
    return vprobas, vlabels


def flatten_probas3d(probas, labels, ignore=None):
    """
    Flattens predictions in the batch
    """
    if probas.dim() == 4:
        # assumes output of a sigmoid layer
        B, D, H, W = probas.size()
        probas = probas.view(B, 1, D, H, W)
    B, C, D, H, W = probas.size()
    probas = (
        probas.permute(0, 2, 3, 4, 1).contiguous().view(-1, C)
    )  # B * H * W, C = P, C
    labels = labels.view(-1)
    if ignore is None:
        return probas, labels
    valid = labels != ignore
    vprobas = probas[valid.nonzero().squeeze()]
    vlabels = labels[valid]
    return vprobas, vlabels


def xloss(logits, labels, ignore=None):
    """
    Cross entropy loss
    """
    return F.cross_entropy(logits, Variable(labels), ignore_index=255)


# --------------------------- HELPER FUNCTIONS ---------------------------
def isnan(x):
    return x != x


def mean(l, ignore_nan=False, empty=0):  # noqa: E741
    """
    nanmean compatible with generators.
    """
    l = iter(l)  # noqa: E741
    if ignore_nan:
        l = ifilterfalse(isnan, l)  # noqa: E741
    try:
        n = 1
        acc = next(l)
    except StopIteration:
        if empty == "raise":
            raise ValueError("Empty mean")
        return empty
    for n, v in enumerate(l, 2):
        acc += v
    if n == 1:
        return acc
    return acc / n
